pub trait Alignment: Copy {}

impl Alignment for super::A2 {}
impl Alignment for super::A4 {}
impl Alignment for super::A8 {}
impl Alignment for super::A16 {}
impl Alignment for super::A32 {}
impl Alignment for super::A64 {}
